{- |
    Module      :  $Header$
    Description :  Interface for reading and manipulating FlatCurry source code
    Copyright   :  (c) 2014 Björn Peemöller
    License     :  BSD-3-clause

    Maintainer  :  bjp@informatik.uni-kiel.de
    Stability   :  experimental
    Portability :  portable
-}
module Curry.FlatCurry
  ( module Curry.FlatCurry.Type
  , module Curry.FlatCurry.Pretty
  , module Curry.FlatCurry.Files
  ) where

import Curry.FlatCurry.Files
import Curry.FlatCurry.Pretty
import Curry.FlatCurry.Type
